#include <cctype>
#include <iostream>
#include <map>
#include <string>
#include <vector>

#include "dictionary.h"

std::vector<std::string> getWordsByLength(int length);
void gameLoop(std::vector<std::string> &words);
std::string filterWords(std::vector<std::string> *words,
                        std::string guessedLetters);

std::string maskSingleWord(std::string word, std::string guessedLetters);

    int
    main() {
  std::cout << "What word length do you want? (<1 or non-numeric initiates "
               "self-tests) ";
  std::string input;
  std::getline(std::cin, input);
  int length = -1;
  try {
    length = std::stoi(input);
  } catch (...) {
    length = -1;
  }

  if (length < 1) {
    std::cout << "Starting self-tests." << std::endl;

    //--------------------------------------------------------------------------
    // Put calls to your testing code here!
    //--------------------------------------------------------------------------

    std::cout << "Tests complete." << std::endl;
  } else {
    std::vector<std::string> words = getWordsByLength(length);

    std::cout << "There are " << words.size() << " words of length " << length
              << std::endl;

    if (words.size() < 1) {
      std::cout << "Insufficient words to play. Quitting.\n";
    } else {
      gameLoop(words);
    }
  }

  return 0;
}

// Returns a vector of all words of a fixed length
std::vector<std::string> getWordsByLength(int length) {
  std::vector<std::string> words;

  for (const std::string &i : FULL_WORDLIST) {
    if (i.length() == length) {
      words.push_back(i);
    }
  }

  return words;
}

void gameLoop(std::vector<std::string> &words) {
  bool gameOver = false;
  int guessesRemaining = 5;

  std::string guessedLetters = "";
  std::string hint = "";

  // Create initial hint.
  for (unsigned i = 0; i < words[0].length(); ++i)
    hint += "_";

  while (!gameOver) {
    // Display game status
    std::cout << "------------------------------------------\n";

    std::cout << "The possible words are: ";
    for (const auto &i : words) {
      std::cout << i << " ";
    }
    std::cout << "\n";

    std::cout << "There are " << words.size() << " words still possible.\n"
              << "You have " << guessesRemaining << " incorrect guesses left.\n"
              << "Current hint: " << hint << "\n"
              << "Guessed letters: " << guessedLetters << std::endl;

    // Prompt for guess
    char guess = '\0';
    for (;;) {
      std::cout << "What letter do you want to guess? ";
      std::string guessInput = "";
      std::getline(std::cin, guessInput);

      // Ensure it is a valid guess
      if (guessInput.length() != 1) {
        std::cout << "Enter exactly one letter. ";
        continue;
      }
      if (!std::isalpha(guessInput[0])) {
        std::cout << "That was not a letter. Enter exactly one letter. ";
        continue;
      }

      if (guessedLetters.find(guessInput[0]) != std::string::npos) {
        std::cout << "Letter already guessed. ";
        continue;
      }

      // It is a valid guess
      guess = guessInput[0];
      guessedLetters += guess;
      break;
    }
    std::cout << "You have guessed '" << guess << "'" << std::endl;

    // Filter word list

    //--------------------------------------------------------------------------
    // Fill in code here. Make sure to update hint as needed.
    //--------------------------------------------------------------------------

    // Check if game over
    if (guessesRemaining < 1) {
      std::cout << "No guesses remain. The word was " << words[0] << std::endl;
      break;
    }

    if (hint.find('_') == std::string::npos) {
      std::cout << "You win! The word was " << hint << std::endl;

      if (words.size() != 1) {
        std::cerr << "ERROR: Remaining word list not of size 1. Have: ";
        for (const std::string &i : words)
          std::cerr << i << "\n";
      }
      break;
    }
  }
}

// Takes in a vector of words, and the guessedLetters. Modifies the vector of
// words and return the hint string consisting only of underscores and the
// letters that have been guessed that are correct
std::string filterWords(std::vector<std::string> *words,
                        std::string guessedLetters) {
  // Partition the words based upon the masked forms.
  std::map<std::string, std::vector<std::string>> partitions;
  //----------------------------------------------------------------------------
  // Code goes here
  //----------------------------------------------------------------------------

  // Find the partition with the largest size
  //----------------------------------------------------------------------------
  // Code goes here
  //----------------------------------------------------------------------------

  // Update the word list
  //----------------------------------------------------------------------------
  // Code goes here. Ensure *words has the correct list
  // Something like: *words = partition[...];
  //----------------------------------------------------------------------------

  return ""// hint corresponding to largest partition;
}

// Returns a string that consists of the characters in the word if they appear
// in guessedLetters, otherwise an underscore. eg for the word "bombastic"
// (which is a word that can't be used enough), and the guessedLetters "eaprb"
// would return "b__ba____"
std::string maskSingleWord(std::string word, std::string guessedLetters) {
  std::string masked = "";

  //----------------------------------------------------------------------------
  // Code goes here
  //----------------------------------------------------------------------------

  return masked;
}
